/*

   rta.cpp

   Copyright (c) Michael Strickland

   GNU General Public License (GPLv3)
   See detailed text in license directory

*/

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cmath>

using namespace std;

#include "rta.h"
#include "paramreader.h"

// ------------------------------------------------------------------------------------------------
//  Parameters (default values below are overridden by values in params.txt)
// ------------------------------------------------------------------------------------------------

int constant_teq = 0; // if equal to 1 then use constant teq otherwise use constant etabar
// only one of the following two is used depending on the value of the constant_teq variable
double teq = 0.25; // teq in fm/c
double fpieb = 1; // 4 Pi eta / S

double t0 = 0.25; // initial time in fm/c
double tf = 10; // final time in fm/c
double T0 = 0.6; // initial temperature in GeV
double xi0 = 0; // initial anisotropy

int num = 800; // number of grid points to use
int npts = 200; // numbers of points to output
int stepsize = 15; // number of steps per iteration block 

// ------------------------------------------------------------------------------------------------
//  Computed values
// ------------------------------------------------------------------------------------------------

double dt = (tf-t0)/num; // time step in fm/c
double sx0 = sqrt(1+xi0); // computed from xi0
double eb = fpieb/4./M_PI; // etabar = eta/S

// ------------------------------------------------------------------------------------------------

double g0 = 1; // degeneracy factor; arbitrary since it's scaled out in all output
double ebfac = 5;  // 5 for "chapman-enskog"; 6 for grad-israel-stewart; 5 is the correct value,
		   // only use the other value when proving stupid referees wrong!

// ------------------------------------------------------------------------------------------------

double* t4List1;
double* t4List2;

inline double H(double y) {
    if (y==1) return 2;
    if (fabs(y)<1) return y*(fabs(y) + asin(sqrt(1-y*y))/sqrt(1-y*y));
    if (fabs(y)>1) return y*(fabs(y) + asinh(sqrt(y*y-1))/sqrt(y*y-1));
    return 0;
}

inline double Hz(double y) {
    if (y==1) return 2./3.;
    if (y==-1) return -2./3.;
    if (fabs(y)<1) return y*y*y*(fabs(y)/(y*y-1) + asin(sqrt(1-y*y))/pow(1-y*y,3./2.));
    if (fabs(y)>1) return y*y*y*(fabs(y)/(y*y-1) - asinh(sqrt(y*y-1))/pow(y*y-1,3./2.));
    return 0;
}

inline double Ht(double y) {
    if (y==1) return 4./3.;
    if (y==-1) return -4./3.;
    if (fabs(y)<1) return -y*(-fabs(y)*sqrt(1-y*y) + (2*y*y-1)*asin(sqrt(1-y*y)))/pow(1-y*y,3./2.);
    if (fabs(y)>1) return y*(-fabs(y)*sqrt(y*y-1) + (2*y*y-1)*asinh(sqrt(y*y-1)))/pow(y*y-1,3./2.);
    return 0;
}

inline double pt(int i) {
    return t0 + i*dt + 0.5*dt;
}

inline int index(double t) {
    if (t==t0) return 0;
    return roundf((t-t0)/dt - 0.5);
}

inline double teqinv(int i) {
    return pow(t4List1[i],0.25)/ebfac/eb/HBARC;
}

inline double D(double t2, double t1) {
    double temp = 0;
    for (int k=index(t1);k<index(t2);k++) {
        if (constant_teq==1)
            temp += 1/teq;
        else
            temp += teqinv(k);
    }
    temp *= dt;
    if (t1==t0) {
        if (constant_teq==1)
            temp += (pt(0)-t0)/teq;
        else
            temp += (pt(0)-t0)*T0/ebfac/eb/HBARC;
    }
    return exp(-temp);
}

inline double Dold(double t2, double t1) {
    return exp((t1-t2)/teq);
}

inline double tKernel(double t,double tp) {
    double tinv = 0;
    if (constant_teq==1)
        tinv = 1/teq;
    else
        tinv = teqinv(index(tp));
    return D(t,tp)*H(tp/t)*tinv/2;
}

inline double t4Init(double t) {
    return pow(T0*pow(t0/t,1./3.),4.);
}

void loadInitT4() {
    for (int i=0;i<num;i++) t4List1[i] = t4Init(pt(i));
}

inline double doInt(int j) {
    double temp = 0;
    for (int i=0;i<j;i++) {
        temp += tKernel(pt(j),pt(i))*t4List1[i];
    }
    temp += tKernel(pt(j),pt(j))*t4List1[j]/2;
    return temp*dt;
}

inline double doStep(int j) {
    double t = pt(j);
    return D(t,t0)*pow(T0,4.)*H(t0/t/sx0)/H(1/sx0) + doInt(j);
}

inline void Evolve(int n) {
    double *tmp;
    for (int i=0;i<n;i++) {
        cout << "--> Iteration " << i << ":  \t";
        for (int j=0;j<num;j++) {
            t4List2[j] = doStep(j);
        }
        cout << "num/2=" << t4List1[((int)num/2)] << "\t num-1=" << t4List1[num-1] << endl;
        // pointer swap
        tmp = t4List1;
        t4List1 = t4List2;
        t4List2 = tmp;
    }
}

inline double nKernel(double t,double tp) {
    double tinv = 0;
    if (constant_teq==1)
        tinv = 1/teq;
    else
        tinv = teqinv(index(tp));
    return D(t,tp)*tp*tinv/t;
}

inline double nInt(int j) {
    double temp = 0;
    for (int i=0;i<j;i++) {
        temp += nKernel(pt(j),pt(i))*pow(t4List1[i],3./4.);
    }
    temp += nKernel(pt(j),pt(j))*pow(t4List1[j],3./4.)/2;
    return temp*dt;
}

inline double nDensity0() {
    double l04 = 2*pow(T0,4.)/H(1/sx0);
    return (2*g0/M_PI/M_PI)*pow(l04,3./4.)/sx0;
}

inline double nDensity(int j) {
    double t = pt(j);
    double l04 = 2*pow(T0,4.)/H(1/sx0);
    return (2*g0/M_PI/M_PI)*(D(t,t0)*pow(l04,3./4.)*t0/t/sx0 + nInt(j));
}

inline double pzKernel(double t,double tp) {
    double tinv = 0;
    if (constant_teq==1)
        tinv = 1/teq;
    else
        tinv = teqinv(index(tp));
    return D(t,tp)*Hz(tp/t)*tinv/2;
}

inline double pzInt(int j) {
    double temp = 0;
    for (int i=0;i<j;i++) {
        temp += pzKernel(pt(j),pt(i))*t4List1[i];
    }
    temp += pzKernel(pt(j),pt(j))*t4List1[j]/2;
    return temp*dt;
}

inline double pParallel0() {
    return (6*g0/M_PI/M_PI)*pow(T0,4.)*Hz(1/sx0)/H(1/sx0);
}

inline double pParallel(int j) {
    double t = pt(j);
    return 6*g0/M_PI/M_PI*(D(t,t0)*pow(T0,4.)*Hz(t0/t/sx0)/H(1/sx0) + pzInt(j));
}

inline double ptKernel(double t,double tp) {
    double tinv = 0;
    if (constant_teq==1)
        tinv = 1/teq;
    else
        tinv = teqinv(index(tp));
    return D(t,tp)*Ht(tp/t)*tinv/2;
}

inline double ptInt(int j) {
    double temp = 0;
    for (int i=0;i<j;i++) {
        temp += ptKernel(pt(j),pt(i))*t4List1[i];
    }
    temp += ptKernel(pt(j),pt(j))*t4List1[j]/2;
    return temp*dt;
}

inline double pPerp0() {
    return (3*g0/M_PI/M_PI)*pow(T0,4.)*Ht(1/sx0)/H(1/sx0);
}

inline double pPerp(int j) {
    double t = pt(j);
    return 3*g0/M_PI/M_PI*(D(t,t0)*pow(T0,4.)*Ht(t0/t/sx0)/H(1/sx0) + ptInt(j));
}

void outputObservables(int idx) {
	fstream out;
	char fname[255];
    // output number density
	sprintf(fname,"output/ndens_%d.dat",idx);
	out.open(fname, ios::out);
    double n0 = nDensity0();
    out << t0 << "\t" << 1 << "\t" << 1 << endl;
	for (int i=0; i<num; i+=num/npts)
        out << pt(i) << "\t" <<  nDensity(i)/n0 << "\t" << t0/pt(i) << endl;
	out.close();
    // output parallel pressure
	sprintf(fname,"output/pz_%d.dat",idx);
	out.open(fname, ios::out);
    double pz0 = pParallel0();
    out << t0 << "\t" << 1 << endl;
	for (int i=0; i<num; i+=num/npts)
        out << pt(i) << "\t" << pParallel(i)/pz0 << endl;
	out.close();
    // output perp pressure
	sprintf(fname,"output/pt_%d.dat",idx);
	out.open(fname, ios::out);
    double pt0 = pPerp0();
    out << t0 << "\t" << 1 << endl;
	for (int i=0; i<num; i+=num/npts)
        out << pt(i) << "\t" << pPerp(i)/pt0 << endl;
	out.close();
    // output pressure ratio
	sprintf(fname,"output/pratio_%d.dat",idx);
	out.open(fname, ios::out);
    out << t0 << "\t" << pz0/pt0 << endl;
	for (int i=0; i<num; i+=num/npts)
        out << pt(i) << "\t" << pParallel(i)/pPerp(i) << endl;
	out.close();
	return;
}

void outputSnapshot(int idx) {
	fstream out;
	char fname[255];
	sprintf(fname,"output/teff_%d.dat",idx);
	out.open(fname, ios::out);
    out << t0 << "\t" << 1 << endl;
	for (int i=0; i<num; i+=num/npts)
        out << pt(i) << "\t" << pow(t4List1[i],1./4.)/T0 << endl;
	out.close();
	return;
}

void solve(int steps) {
    outputSnapshot(0);
    //outputObservables(0);
    for (int i=1;i<=steps;i++) {
        Evolve(stepsize);
        outputSnapshot(i);
    }
    outputObservables(steps);
}

void print_line() {
    const int dwidth = 20;
    for (int i=0;i<5*dwidth;i++) cout << "-"; cout << endl;
    return;
}

int main(int argc, char *argv[])
{
    print_line();
    // read parameters from file and command line
    readParametersFromFile("params.txt",1);
    if (argc>1) {
    	print_line();
        cout << "Parameters from commandline" << endl;
        print_line();
        readParametersFromCommandLine(argc,argv,1);
    }	
    // perform any processing of parameters necessary
    processParameters();

    print_line();
    cout << "==> Allocating Memory" << endl;
    t4List1 = new double[num];
    t4List2 = new double[num];

    print_line();
    cout << "==> Loading Initial Guess" << endl;
    
    loadInitT4();
    
    print_line();
    cout << "==> Computing" << endl;
    print_line();
    
    solve(5);
    
    print_line();
    cout << "==> Freeing Memory" << endl;
    delete[] t4List1;
    delete[] t4List2;

    print_line();

    cout << "==> Done" << endl;
    print_line();
}
